@@ -1,16 +1,17 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2015 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2017 -- leonerd@leonerd.org.uk
 
 package Protocol::IRC;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.11';
+our $VERSION = '0.12';
 
 use Carp;
+use Scalar::Util qw( blessed );
 
 use Protocol::IRC::Message;
 
@@ -27,7 +28,9 @@ allows reading of IRC messages from a string buffer and dispatching them to
 handler methods on its instance.
 
 L<Protocol::IRC::Client> provides an extension to this logic that may be more
-convenient for IRC client implementations.
+convenient for IRC client implementations. Much of the code provided here is
+still useful in client applications, so the reader should be familiar with
+both modules.
 
 =head1 MESSAGE HANDLING
 
@@ -264,14 +267,18 @@ sub incoming_message
    }
 
    if( my $disp = $message->gate_disposition ) {
-      my ( $type, $gate ) = $disp =~ m/^([-+!])(.*)$/;
+      my ( $type, $gate ) = $disp =~ m/^([-+!*])(.*)$/;
       my $effect = ( $type eq "-" ? "more" :
                      $type eq "+" ? "done" :
-                     $type eq "!" ? "fail" : die "TODO" );
+                     $type eq "!" ? "fail" :
+                     $type eq "*" ? ( $hints->{prefix_is_me} ? "done" : undef ) :
+                     die "TODO" );
 
-      $self->invoke( "on_message_gate_${effect}_$gate", $message, $hints ) and $hints->{handled} = 1;
-      $self->invoke( "on_message_gate_$effect", $gate, $message, $hints ) and $hints->{handled} = 1;
-      $self->invoke( "on_message_gate", $effect, $gate, $message, $hints ) and $hints->{handled} = 1;
+      if( defined $effect ) {
+         $self->invoke( "on_message_gate_${effect}_$gate", $message, $hints ) and $hints->{handled} = 1;
+         $self->invoke( "on_message_gate_$effect", $gate, $message, $hints ) and $hints->{handled} = 1;
+         $self->invoke( "on_message_gate", $effect, $gate, $message, $hints ) and $hints->{handled} = 1;
+      }
    }
 
    $self->invoke( "on_message_$command", $message, $hints ) and $hints->{handled} = 1;
@@ -286,14 +293,42 @@ sub incoming_message
 
 =head2 send_message
 
+This method takes arguments in three different forms, depending on their
+number and type.
+
+If the first argument is a reference then it must contain a
+C<Protocol::IRC::Message> instance which will be sent directly:
+
    $irc->send_message( $message )
 
-Sends a message to the peer from the given C<Protocol::IRC::Message>
-object.
+Otherwise, the first argument must be a plain string that gives the command
+name. If the second argument is a hash, it provides named arguments in a form
+similar to L<Protocol::IRC::Message/new_from_named_args>, otherwise the
+remaining arguments must be the prefix string and other positional arguments,
+as plain strings:
+
+   $irc->send_message( $command, { %args } )
 
    $irc->send_message( $command, $prefix, @args )
 
-Sends a message to the peer directly from the given arguments.
+=head3 Named Argument Mangling
+
+For symmetry with incoming message processing, this method applies some
+adjustment of named arguments for convenience of callers.
+
+=over 4
+
+=item *
+
+Callers may define a named argument of C<target>; it will be renamed to
+C<target_name>.
+
+=item *
+
+If a named argument of C<text> is defined and an L</encoder> exists, the
+argument value will be encoded using this encoder.
+
+=back
 
 =cut
 
@@ -305,19 +340,38 @@ sub send_message
 
    if( @_ == 1 ) {
       $message = shift;
+      blessed $message and $message->isa( "Protocol::IRC::Message" ) or
+         croak "Expected an instance of Protocol::IRC::Message";
    }
    else {
-      my ( $command, $prefix, @args ) = @_;
+      my $command = shift;
+      ref $command and
+         croak "Expected \$command to be a plain string";
 
-      if( my $encoder = $self->encoder ) {
-         my $argnames = Protocol::IRC::Message->arg_names( $command );
+      if( @_ == 1 and ref $_[0] ) {
+         my %args = %{ $_[0] };
 
-         if( defined( my $i = $argnames->{text} ) ) {
-            $args[$i] = $encoder->encode( $args[$i] ) if defined $args[$i];
+         $args{target_name} = delete $args{target} if defined $args{target};
+
+         if( defined $args{text} and my $encoder = $self->encoder ) {
+            $args{text} = $encoder->encode( $args{text} );
          }
-      }
 
-      $message = Protocol::IRC::Message->new( $command, $prefix, @args );
+         $message = Protocol::IRC::Message->new_from_named_args( $command, %args );
+      }
+      else {
+         my ( $prefix, @args ) = @_;
+
+         if( my $encoder = $self->encoder ) {
+            my $argnames = Protocol::IRC::Message->arg_names( $command );
+
+            if( defined( my $i = $argnames->{text} ) ) {
+               $args[$i] = $encoder->encode( $args[$i] ) if defined $args[$i];
+            }
+         }
+
+         $message = Protocol::IRC::Message->new( $command, $prefix, @args );
+      }
    }
 
    $self->write( $message->stream_to_line . "\x0d\x0a" );
@@ -671,6 +725,9 @@ sub _on_message_text
 
 =head1 REQUIRED METHODS
 
+As this class is an abstract base class, a concrete implementation must
+provide the following methods to complete it and make it useable.
+
 =cut
 
 =head2 write
