@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2008-2015 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2008-2016 -- leonerd@leonerd.org.uk
 
 package Protocol::IRC::Message;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.11';
+our $VERSION = '0.12';
 
 use Carp;
 our @CARP_NOT = qw( Net::Async::IRC );
@@ -103,6 +103,46 @@ sub new
    return $class->new_with_tags( $_[0], {}, $_[1], @_[2..$#_] );
 }
 
+=head2 new_from_named_args
+
+   $message = Protocol::IRC::Message->new_from_named_args( $command, %args )
+
+Returns a new C<Protocol::IRC::Message> object, initialised from the given
+named argmuents. The argument names must match those required by the given
+command.
+
+=cut
+
+sub new_from_named_args
+{
+   my $class = shift;
+   my ( $command, %args ) = @_;
+
+   my $argnames = $class->arg_names( $command );
+
+   my @args;
+
+   foreach my $name ( keys %$argnames ) {
+      my $idx = $argnames->{$name};
+
+      # Clients don't get to set prefix nick
+      # TODO: servers do
+      next if $idx eq "pn";
+
+      defined( my $value = $args{$name} ) or
+         croak "$command requires a named argmuent of '$name'";
+
+      if( $idx =~ m/^\d+$/ ) {
+         $args[$idx] = $args{$name};
+      }
+      else {
+         die "TODO: not sure what to do with argname idx $idx\n";
+      }
+   }
+
+   return $class->new( $command, undef, @args );
+}
+
 =head2 new_with_tags
 
    $mesage = Protocol::IRC::Message->new_with_tags( $command, \%tags, $prefix, @args )
@@ -375,7 +415,7 @@ my %ARG_NAMES = (
 
 # Misc. named commands
 $ARG_NAMES{$_} = { target_name => 0 } for qw(
-   JOIN LIST NAMES WHO WHOIS WHOWAS
+   LIST NAMES WHO WHOIS WHOWAS
 );
 
 # TODO: 472 ERR_UNKNOWNMODE: <char> :is unknown mode char to me for <channel>
@@ -422,11 +462,18 @@ the result will be split on spaces and stored in an ARRAY ref.
 
 =back
 
+=head2 arg_names (class method)
+
+   $names = Protocol::IRC::Message->arg_names( $command )
+
+This method may also be invoked as a class method by passing in the command
+name or numeric. This allows inspection of what arguments would be required
+or returned before a message object itself is constructed.
+
 =cut
 
 sub arg_names
 {
-   # Usage: Class->arg_names($command) or $self->arg_names()
    my $command;
 
    if( ref $_[0] ) {
@@ -516,6 +563,11 @@ it.
 
 Completes the gate with a successful result.
 
+=item *GATE
+
+Completes the gate with a successful result, but only if the nick in the
+message prefix relates to the connection it is received on.
+
 =item !GATE
 
 Completes the gate with a failure result.
@@ -541,8 +593,10 @@ Paul Evans <leonerd@leonerd.org.uk>
 local $_;
 while( <DATA> ) {
    chomp;
-   my ( $numname, $args, $gating ) = split m/\s*\|\s*/, $_ or next;
-   my ( $num, $name ) = split m/=/, $numname;
+   m/^\s*#/ and next; # ignore comments
+
+   my ( $cmdname, $args, $gating ) = split m/\s*\|\s*/, $_ or next;
+   my ( $cmd, $name ) = split m/=/, $cmdname;
 
    my $index = 0;
    my %args = map {
@@ -555,9 +609,9 @@ while( <DATA> ) {
       }
    } split m/,/, $args;
 
-   $NUMERIC_NAMES{$num} = $name;
-   $ARG_NAMES{$num} = \%args;
-   $GATE_DISPOSITIONS{$num} = $gating if defined $gating;
+   $NUMERIC_NAMES{$cmd} = $name;
+   $ARG_NAMES{$cmd} = \%args;
+   $GATE_DISPOSITIONS{$cmd} = $gating if defined $gating;
 }
 close DATA;
 
@@ -572,6 +626,8 @@ close DATA;
 #   http://www.alien.net.au/irc/irc2numerics.html
 
 __DATA__
+JOIN | 0=target_name | *join
+
 001=RPL_WELCOME         | text
 002=RPL_YOURHOST        | text
 003=RPL_CREATED         | text
@@ -632,7 +688,7 @@ __DATA__
 
 401=ERR_NOSUCHNICK              | target_name,text
 402=ERR_NOSUCHSERVER            | server_name,text
-403=ERR_NOSUCHCHANNEL           | target_name,text
+403=ERR_NOSUCHCHANNEL           | target_name,text | !join
 404=ERR_CANNOTSENDTOCHAN        | target_name,text
 405=ERR_TOOMANYCHANNELS         | target_name,text
 406=ERR_WASNOSUCHNICK           | target_name,text
@@ -649,12 +705,28 @@ __DATA__
 
 467=ERR_KEYSET                  | target_name,text
 
-471=ERR_CHANNELISFULL           | target_name,text
-473=ERR_INVITEONLYCHAN          | target_name,text
-474=ERR_BANNEDFROMCHAN          | target_name,text
-475=ERR_BADCHANNELKEY           | target_name,text
-476=ERR_BADCHANMASK             | target_name,text
-477=ERR_NEEDREGGEDNICK          | target_name,text
+471=ERR_CHANNELISFULL           | target_name,text | !join
+473=ERR_INVITEONLYCHAN          | target_name,text | !join
+474=ERR_BANNEDFROMCHAN          | target_name,text | !join
+475=ERR_BADCHANNELKEY           | target_name,text | !join
+476=ERR_BADCHANMASK             | target_name,text | !join
+477=ERR_NEEDREGGEDNICK          | target_name,text 
 478=ERR_BANLISTFULL             | target_name,text
 
 482=ERR_CHANOPRIVSNEEDED        | target_name,text
+
+# WATCH related - see
+#   http://archives.darenet.org/irc/misc/irc-docs-master/misc/irc-documentation-jilles/reference/draft-meglio-irc-watch-00.txt
+598=RPL_GONEAWAY       | target_name,ident,host,timestamp,text
+599=RPL_NOTAWAY        | target_name,ident,host,timestamp,text
+600=RPL_LOGON          | target_name,ident,host,timestamp,text
+601=RPL_LOGOFF         | target_name,ident,host,timestamp,text
+602=RPL_WATCHOFF       | target_name,ident,host,timestamp,text
+603=RPL_WATCHSTAT      | text
+604=RPL_NOWON          | target_name,ident,host,timestamp,text
+605=RPL_NOWOFF         | target_name,ident,host,timestamp,text
+606=RPL_WATCHLIST      | 1@=nicks
+607=RPL_ENDOFWATCHLIST | text
+609=RPL_NOWISAWAY      | target_name,ident,host,timestamp,text
+
+671=RPL_WHOISSECURE | target_name,text | -whois
